import unittest
from unittest import mock
from unittest.mock import Mock, mock_open

from pycrunch.session.auto_configuration import AutoConfiguration


class TestAutoConfigDecision(unittest.TestCase):
    def setUp(self):
        self.config_file_mock = Mock()
        self.create_function_mock = Mock()

    def test_if_config_exists_do_not_create_new(self):
        self.when_configuration_file_already_on_disk()
        sut = self.create_sut()

        sut.ensure_configuration_exist()

        self.create_function_mock.assert_not_called()

    def test_if_no_config_should_call_create_default(self):
        self.when_file_does_not_exist()
        sut = self.create_sut()

        sut.ensure_configuration_exist()

        self.create_function_mock.assert_called_once()

    def create_sut(self):
        sut = AutoConfiguration(self.config_file_mock)
        sut.create_default_configuration_file = self.create_function_mock
        return sut

    def when_file_does_not_exist(self):
        self.set_config_file_existence(False)

    def when_configuration_file_already_on_disk(self):
        self.set_config_file_existence(True)

    def set_config_file_existence(self, val):
        self.config_file_mock.exists.return_value = val


class TestAutoConfigCreation(unittest.TestCase):
    def setUp(self):
        self.config_file_mock = Mock()
        self.create_function_mock = Mock()

    def test_writes_to_file_at_path(self):
        mocked_open = mock_open()
        with mock.patch('io.open', mocked_open) as io_mock:
            sut = AutoConfiguration(self.config_file_mock)
            sut.create_default_configuration_file()
            handle = io_mock()
            default_config_file ='''# documentation https://pycrunch.com/docs/configuration-file
engine:
    runtime: pytest
    enable-web-ui: false
    load-pytest-plugins: false
'''
            handle.write.assert_called_once_with(default_config_file)




import io
import pickle
import struct
from unittest import TestCase
from unittest.mock import Mock

from pycrunch.scheduling.messages import HandshakeMessage
from pycrunch.networking.server_protocol import TestRunnerServerProtocol
from pycrunch.scheduling.sheduled_task import TestRunPlan
from pycrunch.tests.test_scheduling import generate_tests


class ServerProtocolTests(TestCase):
    def setUp(self) -> None:
        tests = generate_tests(3)
        self.plan = []
        self.plan.append(TestRunPlan(tests, id='1'))
        self.plan.append(TestRunPlan(tests, id='2'))
        self.plan.append(TestRunPlan(tests, id='3'))

    def test_single_message(self):
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)

        print(struct.calcsize("i"))
        handshake = HandshakeMessage('42')
        msg = pickle.dumps(handshake)
        header_bytes = struct.pack("i", len(msg))
        sut.feed_datagram(header_bytes + msg)
        msg = list(sut.message_queue.queue)
        assert msg[0].kind == 'handshake'
        assert msg[0].task_id == '42'

    def test_multiple_messages_in_one_payload(self):
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)

        handshake = HandshakeMessage('43')
        msg = pickle.dumps(handshake)
        header_bytes = struct.pack("i", len(msg))

        handshake = HandshakeMessage('44')
        msg2 = pickle.dumps(handshake)
        header_bytes2 = struct.pack("i", len(msg))

        sut.feed_datagram(header_bytes + msg + header_bytes2 + msg2)
        message_queue = list(sut.message_queue.queue)
        assert len(message_queue) == 2
        assert message_queue[0].task_id == '43'
        assert message_queue[1].task_id == '44'

    def test_message_split_into_two_packets(self):
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)

        handshake = HandshakeMessage('43')
        msg = pickle.dumps(handshake)
        length_of_message = len(msg)
        header_bytes = struct.pack("i", length_of_message)
        half = round(length_of_message / 2)
        print(length_of_message)
        msg_1 = msg[:half]
        msg_2 = msg[half:]

        print(msg_1)
        print(msg_2)

        recovered_bytes = msg_1 + msg_2
        msg_recovered = pickle.loads(recovered_bytes)
        print(msg_recovered)
        print(msg_recovered.task_id)
        sut.feed_datagram(header_bytes + msg_1)
        sut.feed_datagram(msg_2)
        message_queue = list(sut.message_queue.queue)
        assert len(message_queue) == 1

        pass

    def test_message_split_into_three_packets_with_overrun_message(self):
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)

        handshake = HandshakeMessage('43')
        msg = pickle.dumps(handshake)
        length_of_message = len(msg)
        header_bytes = struct.pack("i", length_of_message)
        half = round(length_of_message / 2)
        print(length_of_message)
        msg_1 = msg[:half]
        msg_2 = msg[half:]

        print(msg_1)
        print(msg_2)

        handshake2 = HandshakeMessage('44')
        msg4_bytes = pickle.dumps(handshake2)
        length_of_message2 = len(msg4_bytes)
        header_bytes2 = struct.pack("i", length_of_message2)
        half2 = round(length_of_message2 / 2)
        msg_3 = msg4_bytes[:half]
        msg_4 = msg4_bytes[half:]

        recovered_bytes = msg_1 + msg_2
        msg_recovered = pickle.loads(recovered_bytes)
        print(msg_recovered)
        print(msg_recovered.task_id)
        sut.feed_datagram(header_bytes + msg_1)
        sut.feed_datagram(msg_2 + header_bytes2)
        sut.feed_datagram(msg_3)
        sut.feed_datagram(msg_4)
        message_queue = list(sut.message_queue.queue)
        assert len(message_queue) == 2
        assert message_queue[0].task_id == '43'
        assert message_queue[1].task_id == '44'

        pass

    def test_message_header_split_into_packets(self):
        # -> {length}{buffer}{length}{buffer}
        # <- {length}{buffer}{len
        # <-  gth}{buffer}
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)

        handshake = HandshakeMessage('43')
        msg = pickle.dumps(handshake)
        length_of_message = len(msg)
        header_bytes = struct.pack("i", length_of_message)
        half = round(length_of_message / 2)
        print(length_of_message)
        msg_1 = msg[:half]
        msg_2 = msg[half:]

        print(msg_1)
        print(msg_2)

        handshake2 = HandshakeMessage('44')
        msg4_bytes = pickle.dumps(handshake2)
        length_of_message2 = len(msg4_bytes)
        header_bytes2 = struct.pack("i", length_of_message2)
        half2 = round(length_of_message2 / 2)
        msg_3 = msg4_bytes[:half]
        msg_4 = msg4_bytes[half:]

        recovered_bytes = msg_1 + msg_2
        msg_recovered = pickle.loads(recovered_bytes)
        print(msg_recovered)
        print(msg_recovered.task_id)
        sut.feed_datagram(header_bytes + msg_1)
        sut.feed_datagram(msg_2 + header_bytes2[:2])
        sut.feed_datagram(header_bytes2[2:] + msg_3)
        sut.feed_datagram(msg_4)
        message_queue = list(sut.message_queue.queue)
        assert len(message_queue) == 2
        assert message_queue[0].task_id == '43'
        assert message_queue[1].task_id == '44'

        pass

    def create_sut(self):
        sut = TestRunnerServerProtocol(self.plan, Mock(), Mock())
        sut.process_messages = Mock()
        return sut

    def test_random_bytes(self):
        # -> {length}{buffer}{length}{buffer}
        # <- {length}{buffer}{len
        # <-  gth}{buffer}
        sut = self.create_sut()
        transport = Mock()
        sut.connection_made(transport)
        stream = io.BytesIO()

        stream.write(self.transport_message_bytes(HandshakeMessage('43')))
        stream.write(self.transport_message_bytes(HandshakeMessage('44')))
        stream.write(self.transport_message_bytes(HandshakeMessage('45')))
        stream.write(self.transport_message_bytes(HandshakeMessage('45')))
        stream.write(self.transport_message_bytes(HandshakeMessage('45')))
        stream.write(self.transport_message_bytes(HandshakeMessage('49')))
        buf = stream.getbuffer()
        print(f'lll {len(buf)}')
        read_stream = io.BytesIO(buf.tobytes())

        while True:
            current_byte = read_stream.read(3)
            if len(current_byte) == 0:
                break

            sut.feed_datagram(current_byte)

        # sut.feed_datagram(msg_2 + header_bytes2[:2])
        # sut.feed_datagram(header_bytes2[2:] + msg_3)
        # sut.feed_datagram(msg_4)
        message_queue = list(sut.message_queue.queue)
        assert len(message_queue) == 6
        assert message_queue[0].task_id == '43'
        assert message_queue[1].task_id == '44'
        assert message_queue[5].task_id == '49'

        pass

    def transport_message_bytes(self, msg):
        msg_bytes= pickle.dumps(msg)
        length_of_message = len(msg_bytes)
        header_bytes2 = struct.pack("i", length_of_message)
        return header_bytes2+msg_bytes


import logging

from pycrunch.pipeline import execution_pipeline

logger = logging.getLogger(__name__)

"""
  responsible for handling tasks for
    
    - Test run
    - File tracking
"""

async def dispather_thread():
    logger.info('Dispatcher thread -- Start')
    while True:
        logger.debug('Dispatcher thread -- inside event loop, waiting for task...')
        task = await execution_pipeline.get_task()
        logger.debug('Dispatcher thread -- received task...')
        try:
            await task.run()
        except Exception as e:
            logger.exception('Exception in dispatcher_thread, execution will begin from the next task', exc_info=e)





    logger.debug('Dispatcher thread -- End')



import io
import os
from pathlib import Path
from pprint import pprint
from unittest import mock
from unittest.mock import mock_open, create_autospec

from pycrunch.discovery.simple import SimpleTestDiscovery
from pycrunch.session.configuration import Configuration


def test_simple_discovery():
    actual = run_dogfood_discovery()

    found_flag = False
    for t in actual.tests:
        if t.filename.endswith('test_discovery_specs_demo.py'):
            if t.name == 'test_regular':
                found_flag = True
                break

    assert found_flag

def test_dir_walk():
    for x in Path('/Users/gleb/code/PyCrunch/').glob('**/test*.py'):
        print(x.name)


def test_module_with_tests_simple():
    sut = SimpleTestDiscovery('')
    assert sut.is_module_with_tests('tests_simple')
    assert sut.is_module_with_tests('simple_tests')

def test_module_with_tests_nested():
    sut = SimpleTestDiscovery('')
    assert sut.is_module_with_tests('nested.tests_simple')
    assert sut.is_module_with_tests('api_tests.tests')
    assert sut.is_module_with_tests('nested.even_more.tests_simple')


def test_only_methods_are_discovered_not_variables():
    actual = run_dogfood_discovery()
    test_names = list(map(lambda _: _.name, actual.tests))
    assert len(test_names) > 0
    assert 'test_variable' not in test_names

def test_classes_with_unit_tests_are_discoverable():
    actual = run_dogfood_discovery()
    test_names = list(map(lambda _: _.name, actual.tests))
    assert len(test_names) > 0
    assert 'MyClass::test_method1' in test_names
    assert 'MyClass::test_method2' in test_names
    assert 'TestForDummies::test_method1' in test_names
    assert 'TestForDummies::test_method2' in test_names
    assert 'TestForDummies::helper_method' not in test_names


def run_dogfood_discovery():
    root_folder = Path('.')
    current_folder = root_folder.joinpath('pycrunch', 'tests', 'dogfood').absolute()
    sut = SimpleTestDiscovery(str(root_folder.absolute()), Configuration())
    actual = sut.find_tests_in_folder(str(current_folder.absolute()))
    return actual





import asyncio
import sys

import logging
from datetime import datetime, timedelta


logger = logging.getLogger(__name__)

# This is to prevent running engine process that Pycharm is no longer using
# In seconds
disconnection_deadline = 100
significant_sleep_time = timedelta(seconds=disconnection_deadline*2)


class ConnectionWatchdog:
    def __init__(self):
        self.last_wakeup = datetime.now()
        self.connected_clients = 0
        self.initial_connection_established = False

    def connection_established(self):
        logger.info(f'ConnectionWatchdog->connection_established')
        self.connected_clients += 1
        self.initial_connection_established = True

    def connection_lost(self):
        logger.info(f'ConnectionWatchdog->connection_lost')
        self.connected_clients -= 1

    async def watch_client_connection_loop(self):
        logger.info(f'started ConnectionWatchdog->watch_client_connection_loop')

        # if client was not reconnected in 100 seconds deadline - kill engine
        self.last_wakeup = datetime.now()
        while True:
            logger.debug(f'watch_client_connection_loop-> deciding if we need to close engine')
            await asyncio.sleep(disconnection_deadline)
            now = datetime.now()
            time_interval = now - self.last_wakeup
            logger.debug(f'  last wakeup {str(time_interval)} ago')
            logger.debug(f'  at {datetime.now().time()}')
            self.last_wakeup = now
            if time_interval > significant_sleep_time:
                logger.info('watch_client_connection_loop > significant_sleep_time elapsed, will skip next loop')
                # Probably PC wake up from sleep mode.
                # Sleep one more loop before making decision to terminate
                # Allow client plugins more time to reconnect
                continue

            if self.initial_connection_established and self.connected_clients <= 0:
                logger.warning('!! No connection established within reconnection deadline')
                logger.warning('Engine will exit. Please restart it again if you need to.')
                sys.exit(1)


connection_watchdog = ConnectionWatchdog()



from pycrunch.runner.execution_result import ExecutionResult
from pycrunch.session.combined_coverage import combined_coverage
from pycrunch.session.file_map import test_map
import logging

logger = logging.getLogger(__name__)


class TestState:
    def __init__(self, discovered_test, execution_result, pinned):
        self.discovered_test = discovered_test
        self.pinned = pinned
        self.execution_result = execution_result


    def __repr__(self):
        return f'TestState-> {self.discovered_test.fqn}'

class AllTests:
    def __init__(self):
        # fqn -> TestState
        self.tests = dict()

    def test_discovered(self, fqn, discovered_test, is_pinned):
        # todo preserve state
        self.tests[fqn] = TestState(discovered_test, ExecutionResult(), is_pinned)
        combined_coverage.test_did_removed(fqn)

    def test_will_run(self, fqn):
        test_to_be_run = self.tests.get(fqn, None)
        if test_to_be_run is None:
            # removed?? not parsed?
            return
        test_to_be_run.execution_result.run_did_queued()

    def test_did_run(self, fqn, test_run):
        test_to_be_run = self.tests.get(fqn, None)

        #  TestState
        if test_to_be_run is None:
            # this test no longer exists
            return

        test_to_be_run.execution_result = test_run.execution_result

    def pin_test(self, fqn):
        test_to_pin = self.tests.get(fqn, None)
        test_to_pin.pinned = True

    def unpin_test(self, fqn):
        test_to_unpin = self.tests.get(fqn, None)
        test_to_unpin.pinned = False

    def legacy_aggregated_statuses(self):
        # todo rename to status for consistency
        return {fqn: dict(state=test_run_short_info.execution_result.status) for fqn, test_run_short_info in self.tests.items()}


    def collect_by_fqn(self, fqns):
        result = list()
        logger.info(f'Found {len(fqns)} tests for run')
        for fqn in fqns:

            current_test = self.tests[fqn]
            self.log_test_details(current_test)
            result.append(current_test)

        return result

    def log_test_details(self, current_test):
        logger.debug(f'---')
        logger.debug(f'current_test filename is {current_test.discovered_test.filename}')
        logger.debug(f'current_test fqn is {current_test.discovered_test.fqn}')
        logger.debug(f'current_test module is {current_test.discovered_test.module}')
        logger.debug(f'---')

    def get_pinned_tests(self):
        result = set()
        for test in self.tests.values():
            if test.pinned:
                result.add(test.discovered_test.fqn)
        return result

    def discard_tests_not_in_map(self):
        for fqn in list(self.tests):
            test = self.tests[fqn]
            if not test_map.test_exist(test.discovered_test.filename, fqn):
                logger.debug(f'test no longer in file_map {fqn} - Removed')
                del self.tests[fqn]
                combined_coverage.test_did_removed(fqn)

all_tests = AllTests()